from typing import Optional, IO

import click

from lmk.exc import LMKError
from lmk.process.models import Job


class JobError(LMKError):
    """
    Base class for errors originating from monitoring command-line jobs
    """


class ProcessNotAttached(JobError):
    """ """

    def __init__(self) -> None:
        super().__init__("Process not attached")


class JobNotFound(JobError, click.ClickException):
    """
    Error indicating a job wasn't found
    """

    exit_code = 1

    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f"Job not found: {name}")

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class JobNotRunning(JobError, click.ClickException):
    """
    Error indicating that a job is not running for a command it needs to be
    """

    exit_code = 1

    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f"Job is not running: {name}")

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class NotLoggedIn(JobError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self) -> None:
        super().__init__("Not logged in. Run `lmk login` to log in to LMK.")

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class LLDBNotFound(JobError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self) -> None:
        super().__init__("`lldb` executable not found.")

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class LLDBCannotAttach(JobError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self) -> None:
        super().__init__(
            "`lldb` is not allowed to attach to processes, which means "
            "you can't monitor alreay-running processes. Try monitoring "
            "your command with the `lmk run` command if possible, or see "
            "https://docs.lmkapp.dev/docs/cli/running-process for information "
            "on how to allow `lldb` to attach to processes."
        )

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class JobRaisedError(JobError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self, job: Job) -> None:
        self.job = job
        super().__init__(
            f"Job {job.name} encountered error: "
            f"{job.error_type or '<unknown>'}: "
            f"{job.error or '<unknown>'}"
        )

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class JobExitedUnexpectedly(JobError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self, job: Job) -> None:
        self.job = job
        super().__init__(
            f"Job {job.name} exited unexpectedly, unable to retrieve exit code"
        )

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class CannotDetermineShell(LMKError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self) -> None:
        super().__init__("Cannot determine shell")
        self.exit_code = 1

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)


class NoShellJobFound(LMKError, click.ClickException):
    """ """

    exit_code = 1

    def __init__(self, job_id: int) -> None:
        super().__init__(f"No shell job found with ID: {job_id}")
        self.job_id = job_id

    def show(self, file: Optional[IO] = None) -> None:
        click.secho(str(self), fg="red", file=file)



"""
Patches for generated classes
"""
from functools import wraps

from lmk.generated.models.notification_channel_response_payload import (
    NotificationChannelResponsePayload,
)
from lmk.generated.models.email_channel_payload import EmailChannelPayload
from lmk.generated.models.notification_channel_response import (
    NotificationChannelResponse,
)
from lmk.generated.models.text_message_channel_payload import TextMessageChannelPayload


def unpatch():
    pass


def patch():
    """
    This method does not patch globals or do anything dangerous.
    It is patching the __repr__ methods for a few of the generated
    classes from the OpenAPI spec. Patching specific things
    seems more attractive maintaining the classes myself
    """
    global unpatch

    originals = {}

    def patch_method(cls, method):
        original = getattr(cls, method)
        originals.setdefault(cls, {})[method] = original

        def dec(func):
            wrapped = wraps(original)(func)
            setattr(cls, method, wrapped)
            return wrapped

        return dec

    @patch_method(NotificationChannelResponsePayload, "__repr__")
    def create_notification_channel_repr(self):
        return repr(self.actual_instance)

    @patch_method(EmailChannelPayload, "__repr__")
    def email_channel_payload_repr(self):
        fields = []
        for field_name in ["type", "email_address"]:
            fields.append(f"{field_name}={repr(getattr(self, field_name))}")
        return f"{type(self).__name__}({', '.join(fields)})"

    @patch_method(TextMessageChannelPayload, "__repr__")
    def text_message_payload_repr(self):
        fields = []
        for field_name in ["type", "phone_number"]:
            fields.append(f"{field_name}={repr(getattr(self, field_name))}")
        return f"{type(self).__name__}({', '.join(fields)})"

    @patch_method(NotificationChannelResponse, "__repr__")
    def notification_channel_response_repr(self):
        fields = []
        for field_name in ["name", "payload", "notification_channel_id", "is_default"]:
            fields.append(f"{field_name}={repr(getattr(self, field_name))}")
        return f"{type(self).__name__}({', '.join(fields)})"

    def unpatch():
        for cls, methods in originals.items():
            for name, value in methods.items():
                setattr(cls, name, value)



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel, Field, StrictStr

class AppResponse(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    app_id: StrictStr = Field(..., alias="appId")
    name: StrictStr = ...
    icon_url: StrictStr = Field(..., alias="iconUrl")
    description: StrictStr = ...
    intro_warning_markdown: Optional[StrictStr] = Field(None, alias="introWarningMarkdown")
    created_at: StrictStr = Field(..., alias="createdAt")
    last_updated_at: StrictStr = Field(..., alias="lastUpdatedAt")
    __properties = ["appId", "name", "iconUrl", "description", "introWarningMarkdown", "createdAt", "lastUpdatedAt"]

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AppResponse:
        """Create an instance of AppResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> AppResponse:
        """Create an instance of AppResponse from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return AppResponse.parse_obj(obj)

        _obj = AppResponse.parse_obj({
            "app_id": obj.get("appId"),
            "name": obj.get("name"),
            "icon_url": obj.get("iconUrl"),
            "description": obj.get("description"),
            "intro_warning_markdown": obj.get("introWarningMarkdown"),
            "created_at": obj.get("createdAt"),
            "last_updated_at": obj.get("lastUpdatedAt")
        })
        return _obj




import asyncio
import dataclasses as dc
import enum
import logging
import time
from datetime import datetime
from typing import Optional, cast

from blinker import signal
from IPython import get_ipython
from IPython.core.interactiveshell import (
    InteractiveShell,
    ExecutionInfo,
    ExecutionResult,
)

from lmk.jupyter.utils import background_ctx


LOGGER = logging.getLogger(__name__)

jupyter_state_changed = signal("juptyer-state-changed")

jupyter_cell_state_changed = signal("jupyter-cell-state-changed")


class IPythonShellState(str, enum.Enum):
    Idle = "idle"
    Running = "running"


@dc.dataclass(frozen=True)
class IPythonCellState:
    execution_count: int
    info: ExecutionInfo
    started_at: datetime
    result: Optional[ExecutionResult]
    finished_at: Optional[datetime]

    def error(self) -> Optional[BaseException]:
        if not self.result:
            return None
        if self.result.error_before_exec:
            return self.result.error_before_exec
        if self.result.error_in_exec:
            return self.result.error_in_exec
        return None


class IPythonHistory:
    def __init__(
        self,
        shell: Optional[InteractiveShell] = None,
    ) -> None:
        if shell is None:
            shell = get_ipython()
        self._state = IPythonShellState.Idle
        self.shell = shell

        self._idle_start: Optional[float] = None
        self._last_cell_state: Optional[IPythonCellState] = None

    @property
    def state(self) -> str:
        return self._state

    @state.setter
    def state(self, value: IPythonShellState) -> None:
        if self._state == value:
            return
        old_value = self._state
        self._state = value
        jupyter_state_changed.send(self, old_state=old_value, new_state=value)

    def _pre_execute(self) -> None:
        with background_ctx(LOGGER, type(self).__name__):
            self._idle_start = None

    def _pre_run_cell(self, info: ExecutionInfo) -> None:
        with background_ctx(LOGGER, type(self).__name__):
            self.state = IPythonShellState.Running
            self._idle_start = None
            execution_count = 0
            if self._last_cell_state:
                execution_count = self._last_cell_state.execution_count + 1
            prev_state = self._last_cell_state
            self._last_cell_state = IPythonCellState(
                execution_count=execution_count,
                info=info,
                started_at=datetime.utcnow(),
                result=None,
                finished_at=None,
            )
            jupyter_cell_state_changed.send(
                self, old_state=prev_state, new_state=self._last_cell_state
            )

    def _post_execute(self) -> None:
        with background_ctx(LOGGER, type(self).__name__):
            self._idle_start = time.time()

    def _post_run_cell(self, result: ExecutionResult) -> None:
        with background_ctx(LOGGER, type(self).__name__):
            self.state = IPythonShellState.Running
            self._idle_start = time.time()
            prev_state = self._last_cell_state
            now = datetime.utcnow()
            started_at = now
            execution_count = result.execution_count
            if self._last_cell_state:
                started_at = self._last_cell_state.started_at
            if self._last_cell_state and execution_count is None:
                execution_count = self._last_cell_state.execution_count
            if execution_count is None:
                execution_count = -1
            self._last_cell_state = IPythonCellState(
                execution_count=execution_count,
                info=cast(ExecutionInfo, result.info),
                started_at=started_at,
                result=result,
                finished_at=now,
            )
            jupyter_cell_state_changed.send(
                self, old_state=prev_state, new_state=self._last_cell_state
            )

    async def main_loop(self) -> None:
        while True:
            cutoff = 1
            if (
                self.state == IPythonShellState.Running
                and self._idle_start
                and time.time() - self._idle_start > cutoff
            ):
                self.state = IPythonShellState.Idle

            await asyncio.sleep(0.1)

    def connect(self) -> None:
        self.shell.events.register("pre_execute", self._pre_execute)
        self.shell.events.register("pre_run_cell", self._pre_run_cell)
        self.shell.events.register("post_execute", self._post_execute)
        self.shell.events.register("post_run_cell", self._post_run_cell)

    def disconnect(self) -> None:
        self.shell.events.unregister("pre_execute", self._pre_execute)
        self.shell.events.unregister("pre_run_cell", self._pre_run_cell)
        self.shell.events.unregister("post_execute", self._post_execute)
        self.shell.events.unregister("post_run_cell", self._post_run_cell)



import asyncio
import io
import json
import logging
import os
import psutil
import shutil
from typing import List, IO, cast

from lmk.utils.asyncio import check_output
from lmk.process import exc
from lmk.process.monitor import ProcessMonitor, MonitoredProcess


LOGGER = logging.getLogger(__name__)

CURRENT_DIR = os.path.dirname(__file__)

MONITOR_SCRIPT_PATH = os.path.join(CURRENT_DIR, "lldb_monitor_script.py")


async def check_lldb() -> None:
    exec_path = shutil.which("lldb")
    if exec_path is None:
        raise exc.LLDBNotFound

    process = await asyncio.create_subprocess_shell(
        "lldb --batch -o r -- echo 1",
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL,
    )
    exit_code = await process.wait()

    if exit_code != 0:
        raise exc.LLDBCannotAttach


async def run_with_lldb(
    argv: List[str], log_file: IO[bytes]
) -> asyncio.subprocess.Process:
    """ """
    LOGGER.debug("Getting lldb interpreter info")

    interpreter_info_str = await check_output(
        ["lldb", "--print-script-interpreter-info"]
    )

    LOGGER.debug("lldb interpreter info: %s", interpreter_info_str)

    interpreter_info = json.loads(interpreter_info_str)

    pythonpath_components = [interpreter_info["lldb-pythonpath"]]
    if os.getenv("PYTHONPATH"):
        pythonpath_components.append(os.environ["PYTHONPATH"])

    pythonpath = ":".join(pythonpath_components)

    process = await asyncio.create_subprocess_exec(
        interpreter_info["executable"],
        *argv,
        stdout=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
        stderr=log_file,
        env={**os.environ, "PYTHONPATH": pythonpath},
    )

    return process


class LLDBMonitoredProcess(MonitoredProcess):
    """ """

    def __init__(
        self,
        process: asyncio.subprocess.Process,
        pid: int,
        command: List[str],
        log_file: IO[bytes],
    ) -> None:
        self.process = process
        self.pid = pid
        self.command = command
        self.log_file = log_file

    async def send_signal(self, signum: int) -> None:
        message = json.dumps({"type": "send_signal", "signal": signum}) + "\n"
        stdin = cast(asyncio.StreamWriter, self.process.stdin)
        stdin.write(message.encode())

    async def wait(self) -> int:
        try:
            stdout = cast(asyncio.StreamReader, self.process.stdout)

            stdout_line = asyncio.create_task(stdout.readline())
            wait_task = asyncio.create_task(self.process.wait())
            while True:
                await asyncio.wait(
                    [stdout_line, wait_task], return_when=asyncio.FIRST_COMPLETED
                )
                if wait_task.done():
                    exit_code = wait_task.result()
                    LOGGER.error("Monitor process exited with code: %s", exit_code)
                    return -1

                if stdout_line.done():
                    line = stdout_line.result()
                    try:
                        message = json.loads(line)
                    except json.JSONDecodeError:
                        LOGGER.exception(
                            "Invalid message from LLDB monitor process: %s", line
                        )
                    else:
                        if message.get("type") == "exit":
                            return message["exit_code"]
                        LOGGER.warn(
                            "Unhandled message from monitor process: %s",
                            message.get("type"),
                        )
                    stdout_line = asyncio.create_task(stdout.readline())
        finally:
            self.log_file.close()


class LLDBProcessMonitor(ProcessMonitor):
    def __init__(self, pid: int) -> None:
        self.pid = pid

    async def attach(
        self, output_path: str, log_path: str, log_level: str
    ) -> MonitoredProcess:
        log_file = open(log_path, "ab+", buffering=0)

        with open(output_path, "wb+") as f:
            pass

        process = await run_with_lldb(
            [MONITOR_SCRIPT_PATH, "-l", log_level, str(self.pid), output_path],
            log_file,
        )
        LOGGER.debug("Created lldb process with pid %d", process.pid)

        stdout = cast(asyncio.StreamReader, process.stdout)
        wait_task = asyncio.create_task(process.wait())
        stdout_task = asyncio.create_task(stdout.readline())

        await asyncio.wait(
            [wait_task, stdout_task], return_when=asyncio.FIRST_COMPLETED
        )

        if wait_task.done():
            raise RuntimeError(
                "LLDB attach process exited with code %d", wait_task.result()
            )

        stdout_msg = json.loads(await stdout_task)
        if stdout_msg["type"] != "attached":
            process.kill()
            raise RuntimeError(
                "LLDB attach process gave unexpected response: %s", stdout_msg
            )

        LOGGER.info("Process attached")

        command = psutil.Process(self.pid).cmdline()

        return LLDBMonitoredProcess(process, self.pid, command, log_file)



# coding: utf-8

# flake8: noqa

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from lmk.generated.api.app_api import AppApi
from lmk.generated.api.event_api import EventApi
from lmk.generated.api.headless_auth_api import HeadlessAuthApi
from lmk.generated.api.notification_api import NotificationApi
from lmk.generated.api.oauth_api import OauthApi
from lmk.generated.api.session_api import SessionApi

# import ApiClient
from lmk.generated.api_client import ApiClient
from lmk.generated.configuration import Configuration
from lmk.generated.exceptions import OpenApiException
from lmk.generated.exceptions import ApiTypeError
from lmk.generated.exceptions import ApiValueError
from lmk.generated.exceptions import ApiKeyError
from lmk.generated.exceptions import ApiAttributeError
from lmk.generated.exceptions import ApiException
# import models into sdk package
from lmk.generated.models.access_token_request import AccessTokenRequest
from lmk.generated.models.access_token_response import AccessTokenResponse
from lmk.generated.models.actor_response import ActorResponse
from lmk.generated.models.app_response import AppResponse
from lmk.generated.models.create_headless_auth_session_request import CreateHeadlessAuthSessionRequest
from lmk.generated.models.create_session_request import CreateSessionRequest
from lmk.generated.models.create_session_request_state import CreateSessionRequestState
from lmk.generated.models.email_channel_payload import EmailChannelPayload
from lmk.generated.models.event_notification_configuration import EventNotificationConfiguration
from lmk.generated.models.event_request import EventRequest
from lmk.generated.models.event_response import EventResponse
from lmk.generated.models.headless_auth_refresh_token_request import HeadlessAuthRefreshTokenRequest
from lmk.generated.models.headless_auth_session_response import HeadlessAuthSessionResponse
from lmk.generated.models.jupyter_session_state import JupyterSessionState
from lmk.generated.models.none_session_state import NoneSessionState
from lmk.generated.models.notification_channel_info import NotificationChannelInfo
from lmk.generated.models.notification_channel_response import NotificationChannelResponse
from lmk.generated.models.notification_channel_response_payload import NotificationChannelResponsePayload
from lmk.generated.models.notification_channels_response import NotificationChannelsResponse
from lmk.generated.models.partial_jupyter_session_state import PartialJupyterSessionState
from lmk.generated.models.partial_none_session_state import PartialNoneSessionState
from lmk.generated.models.partial_process_session_state import PartialProcessSessionState
from lmk.generated.models.process_send_signal_request import ProcessSendSignalRequest
from lmk.generated.models.process_session_state import ProcessSessionState
from lmk.generated.models.session_action_request import SessionActionRequest
from lmk.generated.models.session_action_request_body import SessionActionRequestBody
from lmk.generated.models.session_response import SessionResponse
from lmk.generated.models.text_message_channel_payload import TextMessageChannelPayload
from lmk.generated.models.update_session_request import UpdateSessionRequest
from lmk.generated.models.update_session_request_state import UpdateSessionRequestState



from typing import Optional, Any

from IPython import get_ipython
from IPython.core.magic import Magics, line_cell_magic, magics_class, line_magic

from lmk import exc
from lmk.instance import get_instance
from lmk.jupyter.utils import display_widget
from lmk.jupyter.widget import get_widget


@magics_class
class LMKMagics(Magics):
    @line_magic
    def lmk_widget(self, line, cell=None):
        display_widget()

    @line_cell_magic
    def lmk(self, line, cell=None):
        instance = get_instance()
        if not instance.logged_in():
            raise exc.NotLoggedIn()

        args = {"state": "stop", "immediate": bool(cell)}
        tokens = list(map(str.lower, line.strip().split()))
        for token in tokens:
            if token == "immediate":
                args["immediate"] = True
                continue
            if "=" not in token:
                raise ValueError(f"Malformed option: '{token}'")
            key, value = map(str.strip, token.split("=", 1))
            if key == "on":
                if value not in {"none", "stop", "error"}:
                    raise ValueError(f"Invalid option for 'on': '{value}'")
                args["state"] = value
                continue
            if key == "immediate":
                if value not in {"true", "false", "1", "0"}:
                    raise ValueError(f"Invalid option for 'immediate': '{value}'")
                bool_value = value in {"true", "1"}
                args["immediate"] = bool_value
                continue

            raise ValueError(f"Invalid option: '{token}'")

        widget = get_widget()
        widget.set_monitoring_state(**args)

        if cell:
            exec(cell, None, self.shell.user_ns)


def register_magics(
    shell: Optional[Any] = None,
    raise_on_no_shell: bool = True,
) -> bool:
    if shell is None:
        shell = get_ipython()
    if shell is None and raise_on_no_shell:
        raise RuntimeError("Not in an IPython shell")
    if shell is not None:
        shell.register_magics(LMKMagics)
        return True
    return False



import json
import os
import sys
import subprocess
import sys
from typing import List


def main(argv: List[str]) -> int:
    info_str = subprocess.check_output([
        "lldb",
        "--print-script-interpreter-info"
    ])
    info = json.loads(info_str)

    pythonpath_components = [
        info["lldb-pythonpath"]
    ]
    if os.getenv("PYTHONPATH"):
        pythonpath_components.append(os.getenv("PYTHONPATH"))
    
    pythonpath = ":".join(pythonpath_components)

    process = subprocess.Popen(
        [info["executable"]] + argv,
        env={
            **os.environ,
            "PYTHONPATH": pythonpath
        }
    )
    return process.wait()


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import asyncio
from functools import wraps
from typing import Optional, Callable, Any

import click


def async_group(func: Optional[Callable] = None, **kws) -> click.Group:
    def dec(f):
        @click.group(**kws)
        @wraps(f)
        def wrapper(*args, **kwargs):
            return asyncio.run(f(*args, **kwargs))

        return wrapper

    if func is not None:
        return dec(func)

    return dec  # type: ignore


def async_command(
    group: click.Group, **kws
) -> Callable[[Callable[..., Any]], click.Command]:
    def dec(f):
        @group.command(**kws)
        @wraps(f)
        def wrapper(*args, **kwargs):
            return asyncio.run(f(*args, **kwargs))

        return wrapper

    return dec



import asyncio
import logging
import os
import pty
from typing import List

from lmk.process.monitor import ProcessMonitor, MonitoredProcess
from lmk.utils import wait_for_fd, shlex_join


LOGGER = logging.getLogger(__name__)


class MonitoredChildProcess(MonitoredProcess):
    def __init__(
        self,
        process: asyncio.subprocess.Process,
        command: List[str],
        output_fd: int,
        output_path: str,
    ) -> None:
        self.process = process
        self.command = command
        self.output_fd = output_fd
        self.output_path = output_path

    @property
    def pid(self) -> int:  # type: ignore
        return self.process.pid

    async def send_signal(self, signum: int) -> None:
        self.process.send_signal(signum)

    async def wait(self) -> int:
        output_ready = asyncio.create_task(wait_for_fd(self.output_fd))
        wait = asyncio.create_task(self.process.wait())

        with open(self.output_path, "ab+", buffering=0) as output_file:
            while not wait.done():
                await asyncio.wait(
                    [output_ready, wait], return_when=asyncio.FIRST_COMPLETED
                )

                if output_ready.done():
                    output = os.read(self.output_fd, 1000)
                    output_file.write(output)
                    output_ready = asyncio.create_task(wait_for_fd(self.output_fd))

            output_ready.cancel()
            os.set_blocking(self.output_fd, False)

            # Flush any remaining output
            while True:
                try:
                    output = os.read(self.output_fd, 1000)
                except BlockingIOError:
                    break
                output_file.write(output)

            return wait.result()


class ChildMonitor(ProcessMonitor):
    """ """

    def __init__(self, argv: List[str]) -> None:
        if len(argv) < 1:
            raise ValueError("argv must have length >=1")
        self.argv = argv

    async def attach(
        self,
        output_path: str,
        log_path: str,
        log_level: str,
    ) -> MonitoredChildProcess:
        read_output, write_output = pty.openpty()

        proc = await asyncio.create_subprocess_exec(
            *self.argv,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=write_output,
            stderr=write_output,
            bufsize=0,
            start_new_session=True,
        )
        LOGGER.debug(
            "Created child process: [%s], pid: %d", shlex_join(self.argv), proc.pid
        )
        return MonitoredChildProcess(proc, self.argv, read_output, output_path)



import asyncio
import getpass
import sys

from lmk.processmon.gdb import gdb_monitor_process


async def main() -> None:
    pid = int(sys.argv[1])
    async with gdb_monitor_process(
        pid,
        f"./stdout-{pid}.txt",
        f"./sterr-{pid}.txt"
    ) as proc:
        print("USER", getpass.getuser())
        print("EXIT CODE", await proc.wait())
        print("STDOUT", await proc.stdout.read())
        print("STDERR", await proc.stderr.read())
    

if __name__ == "__main__":
    asyncio.run(main())



# This was originally copied from the source code of the ipynbname package,
# didn't want to add a dependency for this relatively simple functionality
# source: https://github.com/msm1089/ipynbname/blob/master/ipynbname/__init__.py
import asyncio
import json
import logging
import os
import urllib.error
import urllib.request
from itertools import chain
from pathlib import Path
from typing import Generator, Tuple, Union, NoReturn, Optional, Dict, Any

import ipykernel
from blinker import signal
from jupyter_core.paths import jupyter_runtime_dir
from traitlets.config import MultipleInstanceError


LOGGER = logging.getLogger(__name__)

notebook_name_changed = signal("notebook-name-changed")

FILE_ERROR = "Can't identify the notebook {}."
CONN_ERROR = (
    "Unable to access server;\n"
    + "ipynbname requires either no security or token based security."
)


def _list_maybe_running_servers(runtime_dir=None) -> Generator[dict, None, None]:
    """Iterate over the server info files of running notebook servers."""
    if runtime_dir is None:
        runtime_dir = jupyter_runtime_dir()
    runtime_dir = Path(runtime_dir)

    if runtime_dir.is_dir():
        for file_name in chain(
            runtime_dir.glob("nbserver-*.json"),  # jupyter notebook (or lab 2)
            runtime_dir.glob("jpserver-*.json"),  # jupyterlab 3
        ):
            yield json.loads(file_name.read_bytes())


def _get_kernel_id() -> str:
    """Returns the kernel ID of the ipykernel."""
    connection_file = Path(ipykernel.get_connection_file()).stem
    kernel_id = connection_file.split("-", 1)[1]
    return kernel_id


def kernel_id() -> Optional[str]:
    try:
        return _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        return None


def _get_sessions(srv):
    """Given a server, returns sessions, or HTTPError if access is denied.
    NOTE: Works only when either there is no security or there is token
    based security. An HTTPError is raised if unable to connect to a
    server.
    """
    try:
        qry_str = ""
        token = srv["token"]
        if token:
            qry_str = f"?token={token}"
        url = f"{srv['url']}api/sessions{qry_str}"
        with urllib.request.urlopen(url) as req:
            return json.load(req)
    except Exception:
        raise urllib.error.HTTPError(CONN_ERROR)


def find_server_and_session() -> (
    Union[Tuple[Dict[str, Any], Dict[str, Any]], Tuple[None, None]]
):
    """
    Find the server and session for the current Jupyter kernel
    """
    try:
        kernel_id = _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        return None, None  # Could not determine
    for srv in _list_maybe_running_servers():
        try:
            sessions = _get_sessions(srv)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return srv, sess
        except Exception:
            pass  # There may be stale entries in the runtime directory
    return None, None


def notebook_name() -> str:
    """Returns the short name of the notebook w/o the .ipynb extension,
    or raises a FileNotFoundError exception if it cannot be determined.
    """
    from lmk.jupyter.colab import colab_support_enabled

    _, session = find_server_and_session()
    if session is None:
        raise FileNotFoundError(FILE_ERROR.format("name"))

    if colab_support_enabled():
        session_name = session.get("notebook", {}).get("name")
        if session_name:
            return session_name

    path = session.get("notebook", {}).get("path")
    if path is None:
        raise FileNotFoundError(FILE_ERROR.format("name"))

    return os.path.basename(path)


class NotebookInfoWatcher:
    """ """

    def __init__(self, poll_interval: float = 30.0) -> None:
        self.poll_interval = poll_interval
        self.notebook_name: Optional[str] = None

    def refresh(self) -> None:
        current_name = notebook_name()
        if current_name == self.notebook_name:
            return
        self.notebook_name, old_name = current_name, self.notebook_name
        notebook_name_changed.send(self, old_value=old_name, new_value=current_name)

    async def main_loop(self) -> NoReturn:
        loop = asyncio.get_running_loop()

        while True:
            try:
                await loop.run_in_executor(None, self.refresh)
            except Exception:
                LOGGER.exception("Error getting notebook name")
            await asyncio.sleep(self.poll_interval)



import asyncio
import contextlib
import inspect
import logging
import signal
import sys
import time
from functools import partial, wraps
from typing import Optional, Awaitable, Any, AsyncGenerator, List, Callable, Dict

from lmk.utils.os import socket_exists


LOGGER = logging.getLogger(__name__)


async def shutdown_loop(
    loop: asyncio.AbstractEventLoop,
    timeout: Optional[float] = None,
    cancel_running: bool = True,
) -> bool:
    current_task = asyncio.current_task()
    tasks = [task for task in asyncio.all_tasks(loop) if task is not current_task]

    if not tasks:
        return True

    if cancel_running:
        for task in tasks:
            task.cancel()

    all_done_coro = asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
    if timeout is None:
        await all_done_coro
        return True

    try:
        await asyncio.wait_for(all_done_coro, timeout)
    except TimeoutError:
        return False

    return True


@contextlib.contextmanager
def loop_ctx(loop: asyncio.AbstractEventLoop):
    try:
        yield
    finally:
        loop.run_until_complete(shutdown_loop(loop))
        loop.close()


async def shutdown_process(
    process: asyncio.subprocess.Process,
    soft_timeout: Optional[float] = None,
    hard_timeout: Optional[float] = None,
) -> int:
    """ """
    start = time.time()

    wait = asyncio.shield(process.wait())
    # Check if the process is already finished first
    try:
        return await asyncio.wait_for(wait, 0)
    except asyncio.TimeoutError:
        pass

    process.send_signal(signal.SIGINT)
    try:
        return await asyncio.wait_for(wait, soft_timeout)
    except (TimeoutError, asyncio.CancelledError):
        pass

    after_soft_timeout = time.time()
    hard_timeout_left = None
    if hard_timeout is not None:
        hard_timeout_left = hard_timeout - (after_soft_timeout - start)

    if hard_timeout_left is None or hard_timeout_left > 0:
        process.terminate()
        try:
            return await asyncio.wait_for(wait, hard_timeout_left)
        except (TimeoutError, asyncio.CancelledError):
            pass

    process.kill()
    return await wait


def asyncio_create_task(
    coro: Awaitable[Any],
    loop: Optional[asyncio.AbstractEventLoop] = None,
    logger: logging.Logger = LOGGER,
) -> asyncio.Task:
    create = asyncio.create_task
    if loop is not None:
        create = loop.create_task

    async def wrapped():
        try:
            await coro
        except Exception:
            logger.exception("Error in %s", coro)

    return create(wrapped())


@contextlib.asynccontextmanager
async def async_signal_handler_ctx(
    signals: List[Any],
    handler: Callable[[int], Awaitable[None]],
) -> AsyncGenerator[None, None]:
    loop = asyncio.get_running_loop()

    tasks = []

    def handle_signal(signum):
        tasks.append(asyncio_create_task(handler(signum), loop))

    for sig in signals:
        loop.add_signal_handler(sig, partial(handle_signal, sig))

    try:
        yield
    finally:
        for sig in signals:
            loop.remove_signal_handler(sig)
        if tasks:
            await asyncio.wait(tasks)


async def wait_for_socket(
    socket_path: str,
    timeout: Optional[float] = None,
    poll_interval: float = 0.1,
) -> None:
    start = time.time()

    while not socket_exists(socket_path) and (
        timeout is None or time.time() - start < timeout
    ):
        await asyncio.sleep(poll_interval)

    if not socket_exists(socket_path):
        raise TimeoutError


async def wait_for_fd(fd: int) -> None:
    loop = asyncio.get_running_loop()
    future: asyncio.Future = asyncio.Future()
    loop.add_reader(fd, future.set_result, None)
    future.add_done_callback(lambda f: loop.remove_reader(fd))
    await future


async def input_async(prompt: str) -> str:
    print(prompt, end="")
    sys.stdout.flush()
    await wait_for_fd(sys.stdin.fileno())
    return sys.stdin.readline()


async def check_output(args: List[str]) -> str:
    kws: Dict[str, Any] = {}
    if sys.version_info < (3, 10):
        kws["loop"] = asyncio.get_running_loop()

    proc = await asyncio.create_subprocess_exec(
        *args,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        **kws,  # type: ignore
    )
    LOGGER.debug("HERE %s", proc)
    stdout, stderr = await proc.communicate()
    LOGGER.debug("HERE2 %s %s", stdout, stderr)
    exit_code = await proc.wait()
    LOGGER.debug("HERE3 %s", exit_code)
    if exit_code == 0:
        return stdout.decode()

    raise CalledProcessError(args, stdout.decode(), stderr.decode(), exit_code)


class CalledProcessError(Exception):
    def __init__(
        self, args: List[str], stdout: str, stderr: str, exit_code: int
    ) -> None:
        self.stdout = stdout
        self.stderr = stderr
        self.exit_code = exit_code
        super().__init__(
            f"Command {args} exited with status {exit_code}.\n"
            f"Stdout:\n{stdout}\n\n"
            f"Stderr:\n{stderr}"
        )


def exponential_backoff(
    failures: int, base: float = 0.5, exponent: float = 2.0
) -> float:
    return base * failures**exponent


def simple_retry_rule(
    max_attempts: Optional[int] = 4,
    backoff: Callable[[int], float] = exponential_backoff,
):
    def rule(failures: int, error: Exception):
        if max_attempts is not None and failures >= max_attempts:
            return None
        return backoff(failures)

    return rule


RetryRule = Callable[[int, Exception], Optional[float]]


def async_retry(func: Optional[Callable] = None, *, rule: Optional[RetryRule] = None):
    if rule is None:
        rule = simple_retry_rule()

    def dec(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            failures = 0
            while True:
                try:
                    return await f(*args, **kwargs)
                except Exception as err:
                    LOGGER.debug("Error in async_retry", exc_info=True)
                    failures += 1
                    sleep_for = rule(failures, err)
                    if sleep_for is None:
                        raise
                    await asyncio.sleep(sleep_for)

        return wrapper

    if func is None:
        return dec

    return dec(func)


def async_callback(
    func: Optional[Callable[..., Any]] = None,
    *,
    loop: Optional[asyncio.AbstractEventLoop] = None,
):
    def dec(f):
        def done_cb(future):
            try:
                future.result()
            except Exception:
                LOGGER.exception(
                    "%s handler: %s raised exception", signal, getattr(f, "__name__", f)
                )

        @wraps(f)
        def handler(*args, **kwargs):
            use_loop = loop
            if use_loop is None:
                use_loop = asyncio.get_event_loop()

            result = f(*args, **kwargs)
            if inspect.isawaitable(result):
                task = use_loop.create_task(result)
                task.add_done_callback(done_cb)
                return task

            return result

        return handler

    if func is None:
        return dec

    return dec(func)


def asyncio_event(loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Event:
    """
    The `loop` parameter was removed in python 3.10
    """
    kws: Dict[str, Any] = {}
    if sys.version_info < (3, 10):
        kws["loop"] = loop or asyncio.get_event_loop()
    return asyncio.Event(**kws)


def asyncio_lock(loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Lock:
    """
    The `loop` parameter was removed in python 3.10
    """
    kws: Dict[str, Any] = {}
    if sys.version_info < (3, 10):
        kws["loop"] = loop or asyncio.get_event_loop()
    return asyncio.Lock(**kws)


def asyncio_queue(
    maxsize: int = 0, loop: Optional[asyncio.AbstractEventLoop] = None
) -> asyncio.Queue:
    """
    The `loop` parameter was removed in python 3.10
    """
    kws: Dict[str, Any] = {}
    if sys.version_info < (3, 10):
        kws["loop"] = loop or asyncio.get_event_loop()
    return asyncio.Queue(maxsize, **kws)


def asyncio_future(loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Future:
    """
    The `loop` parameter was removed in python 3.10
    """
    kws: Dict[str, Any] = {}
    if sys.version_info < (3, 10):
        kws["loop"] = loop or asyncio.get_event_loop()
    return asyncio.Future(**kws)



import asyncio
import atexit
import logging
from concurrent.futures import ThreadPoolExecutor, Executor, Future
from typing import Optional

from lmk.constants import API_URL
from lmk.generated.api_client import ApiClient as DefaultApiClient, Configuration


LOGGER = logging.getLogger(__name__)


class _ExecutorWrapper:
    """
    Wrap an executor to have the same API (at least the parts that the generated API
    client uses) as multiprocessing.pool.ThreadPool -- this is so that rather than
    an ApplyResult it returns a Future that can be wrapped as a coroutine and used with
    asyncio
    """

    def __init__(self, executor: Executor) -> None:
        self.executor = executor

    def apply_async(self, func, args) -> asyncio.Future:
        return asyncio.wrap_future(self.executor.submit(func, *args))

    def close(self) -> None:
        self.executor.shutdown(wait=True)

    def join(self) -> None:
        pass


class ApiClient(DefaultApiClient):
    """
    ApiClient subclass that uses a ThreadPoolExecutor wrapped with _ExecutorWrapper
    rather than a ThreadPool from multiprocessing
    """

    @property
    def pool(self):
        if self._pool is None:
            atexit.register(self.close)
            self._pool = _ExecutorWrapper(ThreadPoolExecutor(self.pool_threads))
        return self._pool


def api_client(
    server_url: Optional[str] = None,
    logger: Optional[logging.Logger] = None,
) -> ApiClient:
    """
    Create an API client instance
    """
    if server_url is None:
        server_url = API_URL
    if logger is None:
        logger = LOGGER

    config = Configuration(host=server_url)
    config.logger = {key: logger for key in config.logger}

    return ApiClient(config)



import importlib
from unittest.mock import patch

import lmk
from lmk.api_client import ApiClient
from lmk.instance import Channels


def test_import_no_requests():
    with patch.object(ApiClient, "call_api") as p:
        r = importlib.reload(lmk)
        r.channels

    p.assert_not_called()


def test_top_level_methods():
    assert isinstance(lmk.channels, Channels)
    assert callable(lmk.notify)
    assert callable(lmk.logged_in)
    assert callable(lmk.login)
    assert callable(lmk.logout)


def test_jupyter():
    assert lmk.jupyter.get_widget()



# Original: https://opensource.apple.com/source/lldb/lldb-179.1/examples/python/process_events.py
import os
import sys
from typing import List

import lldb


# print("HELLO WORLD", lldb, sys.executable, sys.version, sys.argv)


def run_commands(command_interpreter, commands):
    return_obj = lldb.SBCommandReturnObject()
    for command in commands:
        command_interpreter.HandleCommand( command, return_obj )
        if return_obj.Succeeded():
            print("SUCCESS", return_obj.GetOutput())
        else:
            print("FAILED", return_obj)
    

def main(argv: List[str]) -> None:
    target_pid = int(argv[0])

    # Create a new debugger instance
    debugger = lldb.SBDebugger.Create(False)
    debugger.SetAsync(True)
    command_interpreter = debugger.GetCommandInterpreter()

    run_commands(
        command_interpreter,
        [
            "process handle -p true -s false -n true SIGINT SIGTERM SIGCONT",
            "process handle -p true -s true -n true SIGTSTP SIGSTOP"
        ]
    )

    target = debugger.CreateTargetWithFileAndArch(None, lldb.LLDB_ARCH_DEFAULT)
    print("TARGET", target)

    if not target:
        print("TARGET LAUNCH FAILED")
        return
    
    attach_info = lldb.SBAttachInfo(target_pid)

    error = lldb.SBError()
    process = target.Attach(attach_info, error)
    if not process or process.GetProcessID() == lldb.LLDB_INVALID_PROCESS_ID:
        print("PROCESS ATTACH FAILED")
        return

    pid = process.GetProcessID()
    print("PID", pid, process, process.GetState())
    listener = lldb.SBListener("event_listener")

    process.GetBroadcaster().AddListener(listener, lldb.SBProcess.eBroadcastBitStateChanged | lldb.SBProcess.eBroadcastBitStructuredData)
    process.Continue()
    # print("HERE123", listener.WaitForEvent(1, lldb.SBEvent()))
    # try:
    #     process.Continue()
    # except BaseException as err:
    #     print("ERROR", err)
    # print("HERE123", listener.WaitForEvent(1, lldb.SBEvent()))

    stop_idx = 0
    done = False

    while not done:
        event = lldb.SBEvent()
        timeout = 10
        if not listener.WaitForEvent(timeout, event):    
            # timeout waiting for an event
            print("no process event for %u seconds, killing the process..." % timeout)
            done = True
            continue
        
        state = lldb.SBProcess.GetStateFromEvent(event)
        print("STATE", state, event.GetType())
        if lldb.SBProcess.EventIsStructuredDataEvent(event):
            print("IS STRUCTURED")
            data = lldb.SBProcess.GetStructuredDataFromEvent(event)
            print("DATA", data)
        else:
            print("NOT STRUCTURED")
        if state == lldb.eStateStopped:
            thread = process.GetThreadAtIndex(0)
            print("THREAD", thread.GetStopReason())
            print("THREAD 2", thread.GetStopDescription(1000))
            # for thread in process:
            #     print("STOP REASON", thread)
            if stop_idx == 0:
                print("attached to process %u" % (pid))
                # for m in target.modules:
                #     print(m)
                # run_commands(command_interpreter, [])
            else:
                print("process %u stopped" % (pid))
                # run_commands (command_interpreter, [])
            
            continue

        if state == lldb.eStateExited:
            exit_desc = process.GetExitDescription()
            if exit_desc:
                print("process %u exited with status %u: %s" % (pid, process.GetExitStatus(), exit_desc))
            else:
                print("process %u exited with status %u" % (pid, process.GetExitStatus ()))
            # run_commands (command_interpreter, [])
            done = True
            continue
        
        if state == lldb.eStateCrashed:
            print("process %u crashed" % (pid))
            # run_commands(command_interpreter, [])
            done = True
            continue
        
        if state == lldb.eStateDetached:
            print("process %u detached" % (pid))
            done = True
            continue

        if state == lldb.eStateRunning:
            print("process %u resumed" % (pid))
            continue

        if state == lldb.eStateUnloaded:
            print("process %u unloaded, this shouldn't happen" % (pid))
            done = True
            continue

        if state == lldb.eStateConnected:
            print("process connected")
            continue

        if state == lldb.eStateAttaching:
            print("process attaching")
            continue

        if state == lldb.eStateLaunching:
            print("process launching")
            continue
        
        print("UNHANDLED", state)

    # process.Detach()
    print("TERMINATING")
    lldb.SBDebugger.Terminate()


if __name__ == "__main__":
    main(sys.argv[1:])



import json
import os
import uuid
from datetime import date, datetime
from typing import Optional, List, Any

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker

from lmk.process import exc
from lmk.process.models import Base, Job


MISSING: Any = object()


class JobManager:
    """
    Interface for managing and querying job data. Job data is stored in a SQLite database
    """

    def __init__(self, base_path: Optional[str] = None) -> None:
        if base_path is None:
            base_path = os.path.expanduser("~/.lmk")
        self.base_path = base_path
        self.jobs_dir = os.path.join(base_path, "jobs")
        self.engine = create_async_engine(
            f"sqlite+aiosqlite:///{os.path.join(base_path, 'data.db')}"
        )
        self.async_session = async_sessionmaker(
            bind=self.engine, expire_on_commit=False
        )

    async def setup(self) -> None:
        if not os.path.exists(self.base_path):
            os.makedirs(self.base_path)

        if not os.path.exists(self.jobs_dir):
            os.makedirs(self.jobs_dir)

        async with self.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    def _job_dir(self, name: str) -> str:
        return os.path.join(self.jobs_dir, name)

    def pid_file(self, name: str) -> str:
        return os.path.join(self._job_dir(name), "manager.pid")

    def socket_file(self, name: str) -> str:
        return os.path.join(self._job_dir(name), "manager.sock")

    def log_file(self, name: str) -> str:
        return os.path.join(self._job_dir(name), "manager.log")

    def output_file(self, name: str) -> str:
        return os.path.join(self._job_dir(name), "process.log")

    async def create_job(
        self, process_name: Optional[str] = None, notify_on: Optional[str] = None
    ) -> Job:
        if notify_on is None:
            notify_on = "none"

        parts = [process_name] if process_name else []
        parts.append(date.today().strftime("%Y%m%d"))

        async with self.async_session() as session:
            while True:
                job_hash = uuid.uuid4().hex[:8]
                job_id = "-".join(parts + [job_hash])

                existing_job = await session.scalar(
                    sa.select(Job).where(Job.name == job_id)
                )
                if existing_job is not None:
                    continue

                os.makedirs(self._job_dir(job_id))
                obj = Job(
                    name=job_id,
                    notify_on=notify_on,
                )
                session.add(obj)
                await session.commit()
                return obj

    async def get_job(self, name: str, not_started: bool = False) -> Optional[Job]:
        async with self.async_session() as session:
            query = sa.select(Job).where(Job.name == name)
            if not_started:
                query = query.where(Job.started_at == None)
            return await session.scalar(query)

    async def start_job(self, name: str) -> Job:
        job = await self.get_job(name)
        if job is None:
            raise exc.JobNotFound(name)

        async with self.async_session() as session:
            job = await session.merge(job, load=False)
            job.started_at = datetime.utcnow()
            session.add(job)
            await session.commit()

            return job

    async def update_job(
        self,
        name: str,
        pid: int = MISSING,
        command: List[str] = MISSING,
        notify_on: str = MISSING,
        notify_status: str = MISSING,
        channel_id: Optional[str] = MISSING,
        session_id: str = MISSING,
    ) -> Job:
        job = await self.get_job(name)
        if job is None:
            raise exc.JobNotFound(name)

        async with self.async_session() as session:
            job = await session.merge(job, load=False)
            if pid is not MISSING:
                job.pid = pid
            if command is not MISSING:
                job.command = json.dumps(command)
            if notify_on is not MISSING:
                job.notify_on = notify_on
            if notify_status is not MISSING:
                job.notify_status = notify_status
            if channel_id is not MISSING:
                job.channel_id = channel_id
            if session_id is not MISSING:
                job.session_id = session_id

            session.add(job)
            await session.commit()

            return job

    async def end_job(
        self,
        name: str,
        exit_code: int,
        error: Exception = MISSING,
    ) -> Job:
        job = await self.get_job(name)
        if job is None:
            raise exc.JobNotFound(name)

        async with self.async_session() as session:
            job = await session.merge(job, load=False)
            job.exit_code = exit_code
            if error is not MISSING:
                job.error_type = type(error).__name__
                job.error = str(error)
            job.ended_at = datetime.utcnow()

            session.add(job)
            await session.commit()

            return job

    async def list_jobs(self, running_only: bool = False) -> List[Job]:
        query = (
            sa.select(Job).where(Job.started_at != None).order_by(Job.started_at.desc())
        )
        if running_only:
            query = query.where(Job.ended_at == None)

        async with self.async_session() as session:
            return list(await session.scalars(query))



import asyncio
import contextlib
import json
import logging
from typing import Any, Optional, AsyncGenerator

import aiohttp
from blinker import signal

from lmk.utils.blinker import wait_for_signal
from lmk.utils.asyncio import (
    async_retry,
    RetryRule,
    asyncio_queue,
    asyncio_event,
    asyncio_future,
)


LOGGER = logging.getLogger(__name__)


ws_connected = signal("ws-connected")

ws_disconnected = signal("ws-disconnected")

ws_closed = signal("ws-closed")


class WebSocket:
    """
    Wrapper for aiohttp's web socket interface that supports reconnecting on failures
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        url: str,
        retry_rule: Optional[RetryRule] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        **kwargs,
    ) -> None:
        self.session = session
        self.url = url
        self.kwargs = kwargs
        self.retry_rule = retry_rule
        self.loop = loop

        self.queue = asyncio_queue(loop=loop)
        self.ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self.ws_ctx = None
        self.send_task: Optional[asyncio.Task] = None
        self.close_event = asyncio_event(loop=loop)

    def _check_state(self, initialized: bool) -> None:
        if self.ws is None and initialized:
            raise RuntimeError("Context not initialized")
        if self.ws is not None and not initialized:
            raise RuntimeError("Context not initialized")

    async def close(self) -> None:
        self.close_event.set()
        if self.send_task is not None:
            await self.send_task
        if self.ws is not None:
            await self.ws.close()

    async def _setup(self) -> None:
        self._check_state(False)

        @async_retry(rule=self.retry_rule)
        async def init():
            try:
                self.ws_ctx = self.session.ws_connect(self.url, **self.kwargs)
                self.ws = await self.ws_ctx.__aenter__()
                LOGGER.debug("Initialized web socket")
            except:
                LOGGER.exception("Web socket error")
                self.ws_ctx = None
                self.ws = None
                raise
            else:
                ws_connected.send(self)

        await init()

    async def _teardown(self) -> None:
        if self.ws_ctx is None:
            return
        await self.ws_ctx.__aexit__(None, None, None)
        self.ws_ctx = None
        self.ws = None
        ws_disconnected.send(self)

    async def __aenter__(self):
        self.send_task = asyncio.create_task(self._sender())
        await self._setup()
        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        await self.queue.join()

        self.send_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self.send_task

        await self._teardown()
        ws_closed.send(self)

    async def send(self, data: Any) -> None:
        """
        Send a message to the web socket

        :param data: The data to send to the web socket. This must be JSON serializable.
        :type data: Any

        :return: This method does not return anything
        :rtype: None
        """
        future = asyncio_future(loop=self.loop)
        await self.queue.put((data, future))
        await future

    async def _iterate(self) -> AsyncGenerator[Any, None]:
        self._check_state(True)

        close_message: Optional[aiohttp.WSMessage] = None
        while True:
            message = await self.ws.receive()  # type: ignore
            LOGGER.debug("Received message %s", message)
            if message.type == aiohttp.WSMsgType.CLOSED:
                if close_message is None and not self.close_event.is_set():
                    raise WSDisconnected
                break
            if message.type == aiohttp.WSMsgType.CLOSE:
                close_message = message
                continue
            if message.type == aiohttp.WSMsgType.ERROR:
                raise WSConnectionError(message.data)
            if message.type == aiohttp.WSMsgType.TEXT:
                try:
                    decoded = json.loads(message.data)
                except json.JSONDecodeError as err:
                    raise InvalidWSMessage(message.data) from err
                yield decoded
                continue
            if message.type == aiohttp.WSMsgType.CLOSING:
                continue

            LOGGER.warn("Unhandled websocket message type: %s", message)

        if close_message is not None and close_message.data != aiohttp.WSCloseCode.OK:
            raise WSCloseError(close_message.data, close_message.extra)

    async def _iterate_with_retry(self):
        while True:
            try:
                async for item in self._iterate():
                    yield item
                break
            except (WSDisconnected, WSCloseError, WSConnectionError) as error:
                LOGGER.error("WS Disconnected. Reconnecting (%r)", error)
                await self._teardown()
                await self._setup()
            except GeneratorExit:
                break
            except asyncio.CancelledError:
                raise
            except:
                LOGGER.exception("Unexpected error in websocket")
                raise

    async def _sender(self):
        buffer = []

        if self.ws is None:
            connect_task = wait_for_signal(ws_connected, self)
        else:
            connect_task = asyncio_future(loop=self.loop)
            connect_task.set_result(None)

        async def queue_get():
            try:
                return await self.queue.get()
            except RuntimeError as error:
                if str(error) == "Event loop is closed":
                    return None
                raise

        queue_task = asyncio.create_task(queue_get())

        close_task = asyncio.create_task(self.close_event.wait())

        while True:
            LOGGER.debug("Before loop")
            await asyncio.wait(
                [connect_task, queue_task, close_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            LOGGER.debug(
                "Loop; connected: %s, queue: %s", connect_task.done(), queue_task.done()
            )

            if queue_task.done():
                result = queue_task.result()
                if result is not None:
                    buffer.append(queue_task.result())
                queue_task = asyncio.create_task(queue_get())

            while buffer and self.ws is not None:
                item, future = buffer.pop(0)
                await self.ws.send_json(item)
                future.set_result(None)
                self.queue.task_done()

            if connect_task.done():
                connect_task = wait_for_signal(ws_connected, self)

            if close_task.done():
                break

    async def __aiter__(self):
        """
        Iterate asynchronously through messages received by the web socket.

        :return: An asynchronous iterator of messages received from the web socket.
        :rtype: AsyncIterator[Any]
        """
        self._check_state(True)

        async for item in self._iterate_with_retry():
            yield item
            if self.send_task.done():
                self.send_task.result()


class WSError(Exception):
    """ """


class WSDisconnected(WSError):
    """ """

    def __init__(self) -> None:
        super().__init__(f"Web socket disconnected uncleanly")


class WSConnectionError(WSError):
    """ """

    def __init__(self, data: str) -> None:
        super().__init__(f"Web socket error: {data}")


class InvalidWSMessage(WSError):
    """ """

    def __init__(self, data: str) -> None:
        super().__init__(f"Invalid WS message: {data}")


class WSCloseError(WSError):
    """ """

    def __init__(self, code: int, reason: str) -> None:
        super().__init__(
            f"Web socket closed unexpectedly. Code: {code}, " f"reason: {reason}"
        )



import asyncio
import atexit
import contextlib
import enum
import logging
import io
import os
import signal
import threading
import time
import uuid
from datetime import datetime, timedelta, timezone
from typing import Callable, Union, Optional, cast

from ipywidgets import DOMWidget
from traitlets import Unicode, Int, List, Dict, UseEnum

from lmk import exc
from lmk.generated.api.app_api import AppApi
from lmk.generated.models.jupyter_session_state import JupyterSessionState
from lmk.generated.models.session_response import SessionResponse
from lmk.generated.exceptions import ApiException
from lmk.instance import (
    Instance,
    default_instance_changed,
    default_channel_changed,
    access_token_changed,
    server_url_changed,
    channels_fetch_state_changed,
    get_instance,
    ChannelsState,
    Channels,
)
from lmk.jupyter.colab import observe_google_colab_url, colab_support_enabled
from lmk.jupyter.constants import MODULE_NAME, MODULE_VERSION
from lmk.jupyter.history import (
    IPythonHistory,
    jupyter_cell_state_changed,
    jupyter_state_changed,
    IPythonShellState,
)
from lmk.jupyter.notebook_info import (
    NotebookInfoWatcher,
    notebook_name_changed,
    kernel_id,
)
from lmk.jupyter.utils import background_ctx
from lmk.utils.asyncio import loop_ctx, asyncio_event, asyncio_lock, asyncio_queue
from lmk.utils.blinker import wait_for_signal
from lmk.utils.logging import setup_logging
from lmk.utils.ws import WebSocket


LOGGER = logging.getLogger(__name__)

DEFAULT_BASE_PATH = os.path.expanduser("~/.lmk")


def truncate_stream(stream: io.StringIO, max_size: int) -> None:
    value = stream.getvalue()
    if len(value) <= max_size:
        return
    new_size = int(max_size * 0.8)
    stream.truncate(0)
    stream.seek(0)
    stream.write(value[-new_size:])


def ts_millis(dt: Optional[datetime] = None) -> int:
    if dt is None:
        dt = datetime.utcnow()
    return int(dt.replace(tzinfo=timezone.utc).timestamp() * 1000)


def date_from_millis(ts: int) -> datetime:
    return datetime.utcfromtimestamp(ts / 1000.0).replace(tzinfo=timezone.utc)


def format_date(date: datetime) -> str:
    return date.strftime("%Y-%m-%d %H:%M:%S")


class AuthState(str, enum.Enum):
    NeedsAuth = "needs-auth"
    AuthInProgress = "auth-in-progress"
    AuthError = "auth-error"
    Authenticated = "authenticated"


class IPythonCellStateType(str, enum.Enum):
    Running = "running"
    Error = "error"
    Success = "success"
    Cancelled = "cancelled"


class IPythonMonitoringState(str, enum.Enum):
    None_ = "none"
    Error = "error"
    Stop = "stop"


def default_log_path() -> str:
    kernel_uid = kernel_id() or str(uuid.uuid4())
    kernel_path = os.path.join(DEFAULT_BASE_PATH, "jupyter", kernel_uid)
    return os.path.join(kernel_path, "lmk.log")


class LMKWidget(DOMWidget):
    # Metadata needed for jupyter to find the widget
    _model_name = Unicode("LMKModel").tag(sync=True)
    _model_module = Unicode(MODULE_NAME).tag(sync=True)
    _model_module_version = Unicode(MODULE_VERSION).tag(sync=True)
    _view_name = Unicode("LMKView").tag(sync=True)
    _view_module = Unicode(MODULE_NAME).tag(sync=True)
    _view_module_version = Unicode(MODULE_VERSION).tag(sync=True)

    # Notebook metadata
    url = Unicode(None, allow_none=True).tag(sync=True)
    notebook_name = Unicode(None, allow_none=True).tag(sync=True)

    # Auth data
    auth_url = Unicode(None, allow_none=True).tag(sync=True)
    auth_state = UseEnum(AuthState, default_value=AuthState.NeedsAuth).tag(sync=True)
    auth_error = Unicode(None, allow_none=True).tag(sync=True)
    access_token = Unicode(None, allow_none=True).tag(sync=True)
    api_url = Unicode(None, allow_none=True).tag(sync=True)

    # Session data
    session = Dict(None, allow_none=True).tag(sync=True)

    # Jupyter state
    jupyter_state = UseEnum(
        IPythonShellState,
        default_value=None,
        allow_none=True,
    ).tag(sync=True)
    jupyter_execution_num = Int(None, allow_none=True).tag(sync=True)
    jupyter_cell_state = UseEnum(
        IPythonCellStateType, default_value=None, allow_none=True
    ).tag(sync=True)
    jupyter_cell_text = Unicode(None, allow_none=True).tag(sync=True)
    jupyter_cell_started_at = Int(None, allow_none=True).tag(sync=True)
    jupyter_cell_finished_at = Int(None, allow_none=True).tag(sync=True)
    jupyter_cell_error = Unicode(None, allow_none=True).tag(sync=True)

    # Current monitoring state--shared w/ BE
    monitoring_state = Unicode(IPythonMonitoringState.None_.value).tag(sync=True)

    # Params to tweak notification rules--getting notified about something really
    # short where you can obviously see the result would just be annoying
    notify_min_execution = Int(None, allow_none=True).tag(sync=True)
    notify_min_time = Int(None, allow_none=True).tag(sync=True)

    # Metadata about channels
    selected_channel = Unicode(None, allow_none=True).tag(sync=True)
    channels_state = UseEnum(ChannelsState, default_value=ChannelsState.None_).tag(
        sync=True
    )
    channels = List(Dict()).tag(sync=True)

    # Keep a log of sent notifications
    sent_notifications = List(Dict()).tag(sync=True)

    # Sync log level between component and here
    log_level = Unicode("ERROR")
    log_path = Unicode(default_log_path())

    def __init__(self) -> None:
        super().__init__()

        startup_event = threading.Event()

        self.thread = LMKWidgetThread(self, startup_event)
        self.thread.start()

        startup_event.wait(3)

    def set_monitoring_state(
        self,
        state: Union[IPythonMonitoringState, str],
        immediate: bool = False,
    ) -> None:
        if isinstance(state, str):
            state = IPythonMonitoringState(state)
        if state.value == self.monitoring_state:
            return
        self.monitoring_state = state.value
        min_time = ts_millis()
        min_execution = self.jupyter_execution_num
        if not immediate:
            min_time += 2000
            min_execution = (min_execution or 0) + 1

        self.notify_min_execution = min_execution
        self.notify_min_time = min_time

    def set_log_level(self, level: Union[int, str]) -> None:
        if isinstance(level, int):
            level = logging.getLevelName(level)
        self.log_level = cast(str, level)

    def shutdown(self) -> None:
        self.thread.shutdown()


class LMKWidgetThread(threading.Thread):
    """
    Background thread where all of the actual processing for the LMK
    jupyter widget happens
    """

    def __init__(self, widget: LMKWidget, startup_event: threading.Event) -> None:
        super().__init__()
        self.widget = widget
        self.startup_event = startup_event
        self.shutdown_event = threading.Event()
        self.history = IPythonHistory()
        self.info_watcher = NotebookInfoWatcher()
        self.loop = asyncio.new_event_loop()
        self._cancel_auth = asyncio_event(loop=self.loop)
        self._register_shutdown_hook()
        self._setup_logging()

    def _setup_logging(
        self, level: Optional[str] = None, file: Optional[str] = None
    ) -> None:
        if level is None:
            level = self.widget.log_level
        if file is None:
            file = self.widget.log_path

        file_dir = os.path.dirname(file)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir)
        setup_logging(disable_existing=False, level=level, log_file=file)

    def _register_shutdown_hook(self) -> None:
        atexit.register(self.shutdown)

    def shutdown(self) -> None:
        self.shutdown_event.set()
        self.join()

    async def handle_event(self, event):
        if event["name"] == "log_level":
            self._setup_logging(level=event["new"])
        elif event["name"] == "log_path":
            self._setup_logging(file=event["new"])
        elif event["name"] == "selected_channel":
            instance = get_instance()
            instance.default_channel = event["new"]

    async def initiate_auth(self, payload, instance: Instance):
        if self.widget.auth_state == AuthState.AuthInProgress:
            LOGGER.info(f"Skipping auth because it is in progress")
            return

        LOGGER.debug("Initiating auth")

        access_token = None
        try:
            access_token = await instance._get_access_token_async()
        except exc.NotLoggedIn:
            pass
        if access_token and not (payload or {}).get("force"):
            LOGGER.info("Using existing access token")
            self.widget.auth_state = AuthState.Authenticated
            return

        timeout = 300.0
        poll_interval = 1.0
        start = time.time()

        try:
            self.widget.auth_state = AuthState.AuthInProgress
            session = await instance.initiate_auth(async_req=True)  # type: ignore
            self.widget.auth_url = session.authorize_url

            cancellation_wait_task = asyncio.create_task(self._cancel_auth.wait())

            while time.time() - start < timeout and not self._cancel_auth.is_set():
                try:
                    response = await instance.retrieve_auth_token(  # type: ignore
                        session_id=session.session_id, async_req=True
                    )
                except exc.AuthSessionNotComplete:
                    await asyncio.wait(
                        [
                            asyncio.create_task(asyncio.sleep(poll_interval)),
                            cancellation_wait_task,
                        ],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    continue
                else:
                    instance.set_access_token(
                        response.access_token,
                        response.refresh_token,
                        datetime.utcnow() + timedelta(seconds=response.expires_in),
                    )
                    self.widget.auth_state = AuthState.Authenticated
                    self.widget.auth_url = None
                    self.widget.auth_error = None
                    return

            if self._cancel_auth.is_set():
                self._cancel_auth.clear()
                self.widget.auth_url = None
                self.widget.auth_state = (
                    AuthState.Authenticated
                    if instance.logged_in()
                    else AuthState.NeedsAuth
                )
                self.widget.auth_error = None
            else:
                self.widget.auth_state = AuthState.AuthError
                self.widget.auth_url = None
                self.widget.auth_error = "Authentication timed out"
        except Exception as err:
            self.widget.auth_state = AuthState.AuthError
            self.widget.auth_error = f"{type(err).__name__}: {err}"
            raise

    async def cancel_auth(self, payload, instance: Instance) -> None:
        self._cancel_auth.set()

    async def refresh_channels(self, payload, instance: Instance):
        if instance.channels.fetch_state != self.widget.channels_state:
            LOGGER.warn(
                "Widget channels fetch state is not synced properly "
                "with the Instance"
            )
            self.widget.channels_state = instance.channels.fetch_state

        if self.widget.channels_state == ChannelsState.Loading:
            LOGGER.info(f"Skipping channel fetch because it is in progress")
            return

        await instance.channels.fetch(async_req=True, force=True)  # type: ignore

    async def _handle_request_inner(self, request):
        method = request["method"]
        payload = request["payload"]
        instance = get_instance()
        if method == "initiate-auth":
            return await self.initiate_auth(payload, instance)
        if method == "cancel-auth":
            return await self.cancel_auth(payload, instance)
        if method == "refresh-channels":
            return await self.refresh_channels(payload, instance)
        raise ValueError(f"Invalid method {method}")

    async def handle_request(self, request):
        try:
            response = await self._handle_request_inner(request)
            return {
                "request_id": request["request_id"],
                "success": True,
                "payload": response,
            }
        except ApiException as err:
            LOGGER.error(
                "API Request failed with status %d %s: %s",
                err.status,
                err.reason,
                err.body,
            )
            return {
                "request_id": request["request_id"],
                "success": False,
                "error": str(err),
            }
        except Exception as err:
            LOGGER.exception("Error occurred when handling request")
            if not isinstance(request, dict):
                return {
                    "request_id": "<unknown>",
                    "success": False,
                    "error": "Invalid payload",
                }
            return {
                "request_id": request.get("request_id", "<unknown>"),
                "success": False,
                "error": str(err),
            }

    def _handle_done_future(self, future):
        try:
            future.result()
        except Exception:
            LOGGER.exception("Error in callback")

    def _observe_widget(self, loop: asyncio.AbstractEventLoop) -> Callable[[], None]:
        def observe(info):
            LOGGER.debug(
                "Change of `%s` from %r to %r",
                info["name"],
                info["old"],
                info["new"],
            )

            async def handle():
                with background_ctx(LOGGER, type(self).__name__):
                    await self.handle_event(info)

            task = loop.create_task(handle())
            task.add_done_callback(self._handle_done_future)

        self.widget.observe(observe)

        def observe_msg(widget, content, buffers):
            LOGGER.debug("Msg: %s", content)

            async def handle():
                with background_ctx(LOGGER, type(self).__name__):
                    response = await self.handle_request(content)
                    self.widget.send(response)

            task = loop.create_task(handle())
            task.add_done_callback(self._handle_done_future)

        self.widget.on_msg(observe_msg)

        def unobserve():
            self.widget.unobserve(observe)
            self.widget.on_msg(observe_msg, remove=True)

        return unobserve

    async def _send_notification(self) -> None:
        with background_ctx(LOGGER, type(self).__name__):
            instance = get_instance()

            kws = {}
            if self.widget.selected_channel is not None:
                kws["notification_channels"] = [self.widget.selected_channel]

            started = "<unknown>"
            if self.widget.jupyter_cell_started_at is not None:
                started = format_date(date_from_millis(self.widget.jupyter_cell_started_at))

            ended = "<unknown>"
            if self.widget.jupyter_cell_finished_at is not None:
                ended = format_date(date_from_millis(self.widget.jupyter_cell_finished_at))

            if self.widget.jupyter_cell_state == IPythonCellStateType.Error:
                message = (
                    f"Notebook [**{self.widget.notebook_name}**]({self.widget.url}) "
                    f"**failed** during execution **\\[{self.widget.jupyter_execution_num}\\]**:\n"
                    f"```python\n{self.widget.jupyter_cell_text}\n```\n\n"
                    f"Error:\n```\n{self.widget.jupyter_cell_error}\n```\n\n"
                    f"Started: {started}\n\n"
                    f"Ended: {ended}"
                )
            elif self.widget.jupyter_cell_state == IPythonCellStateType.Cancelled:
                message = (
                    f"Notebook [**{self.widget.notebook_name}**]({self.widget.url}) "
                    f"was **cancelled** during execution **\\[{self.widget.jupyter_execution_num}\\]**:\n"
                    f"```python\n{self.widget.jupyter_cell_text}\n```\n\n"
                    f"Started: {started}\n\n"
                    f"Ended: {ended}"
                )
            else:
                message = (
                    f"Notebook [**{self.widget.notebook_name}**]({self.widget.url}) "
                    f"**stopped** after execution **\\[{self.widget.jupyter_execution_num}\\]**:\n"
                    f"```python\n{self.widget.jupyter_cell_text}\n```\n\n"
                    f"Started: {started}\n\n"
                    f"Ended: {ended}"
                )

            response = await instance.notify(message=message, async_req=True, **kws)  # type: ignore
            LOGGER.info("Notification sent. Response: %s", response.to_dict())
            self.widget.sent_notifications = (
                self.widget.sent_notifications + [response.to_dict()]
            )[-100:]

    def _observe_jupyter(self, loop: asyncio.AbstractEventLoop) -> Callable[[], None]:
        def handle_jupyter_state_change(_, old_state, new_state):
            with background_ctx(LOGGER, type(self).__name__):
                self.widget.jupyter_state = new_state
                if (
                    new_state != IPythonShellState.Idle
                    or self.widget.monitoring_state == IPythonMonitoringState.None_
                ):
                    return

                now = ts_millis()
                if (
                    self.widget.notify_min_execution is not None
                    and self.widget.jupyter_execution_num
                    < self.widget.notify_min_execution
                ) or (
                    self.widget.notify_min_time is not None
                    and now < self.widget.notify_min_time
                ):
                    return

                if (
                    self.widget.monitoring_state == IPythonMonitoringState.Error
                    and self.widget.jupyter_cell_state == IPythonCellStateType.Error
                ):
                    task = loop.create_task(self._send_notification())
                    task.add_done_callback(self._handle_done_future)
                elif self.widget.monitoring_state == IPythonMonitoringState.Stop:
                    task = loop.create_task(self._send_notification())
                    task.add_done_callback(self._handle_done_future)
                self.widget.monitoring_state = IPythonMonitoringState.None_

        jupyter_state_changed.connect(handle_jupyter_state_change, sender=self.history)

        def handle_jupyter_cell_state_change(_, old_state, new_state):
            with background_ctx(LOGGER, type(self).__name__):
                # LOGGER.debug("State change %s %s", old_state, new_state)
                self.widget.jupyter_execution_num = new_state.execution_count
                self.widget.jupyter_cell_text = new_state.info.raw_cell
                state = IPythonCellStateType.Running
                error = None
                if new_state.result and new_state.error():
                    error_obj = new_state.error()
                    if isinstance(error_obj, KeyboardInterrupt):
                        state = IPythonCellStateType.Cancelled
                    else:
                        state = IPythonCellStateType.Error
                        error = f"{type(new_state.error()).__name__}: {str(new_state.error())}"
                elif new_state.result:
                    state = IPythonCellStateType.Success
                self.widget.jupyter_cell_state = state
                self.widget.jupyter_cell_started_at = ts_millis(new_state.started_at)
                finished_at = None
                if new_state.finished_at:
                    finished_at = ts_millis(new_state.finished_at)
                self.widget.jupyter_cell_finished_at = finished_at
                self.widget.jupyter_cell_error = error

        jupyter_cell_state_changed.connect(
            handle_jupyter_cell_state_change, sender=self.history
        )

        self.history.connect()

        def unobserve():
            self.history.disconnect()
            jupyter_state_changed.disconnect(
                handle_jupyter_state_change, sender=self.history
            )
            jupyter_cell_state_changed.disconnect(
                handle_jupyter_cell_state_change, sender=self.history
            )

        return unobserve

    def _observe_instance(self, loop: asyncio.AbstractEventLoop) -> Callable[[], None]:
        disconnected = False
        instance = get_instance()

        def connect(instance: Instance) -> None:
            self.widget.selected_channel = instance.default_channel
            self.widget.access_token = instance.access_token
            self.widget.api_url = instance.server_url
            self.widget.auth_url = None

            if instance.access_token:
                self.widget.auth_state = AuthState.Authenticated
            else:
                self.widget.auth_state = AuthState.NeedsAuth

            self.widget.channels_state = instance.channels.fetch_state
            if instance.channels.fetch_state == ChannelsState.Loaded:
                self.widget.channels = [
                    channel.to_dict() for channel in instance.channels
                ]
            else:
                self.widget.channels = []

        connect(instance)

        @default_channel_changed.connect
        def on_channel_changed(sender, old_value, new_value):
            LOGGER.debug("Default channel update: %s %s", old_value, new_value)
            nonlocal instance
            if sender is not instance:
                return
            with background_ctx(LOGGER, type(self).__name__):
                self.widget.selected_channel = new_value

        @access_token_changed.connect
        def on_access_token_changed(sender, old_value, new_value):
            nonlocal instance
            if sender is not instance:
                return

            with background_ctx(LOGGER, type(self).__name__):
                if old_value == new_value:
                    return
                self.widget.auth_state = (
                    AuthState.Authenticated if new_value else AuthState.NeedsAuth
                )
                self.widget.channels_state = ChannelsState.None_
                self.widget.channels = []
                self.widget.selected_channel = None
                self.widget.auth_url = None
                self.widget.access_token = new_value
                if new_value:
                    LOGGER.info("Initiating channel fetch")
                    task = loop.create_task(self.refresh_channels({}, instance))
                    task.add_done_callback(self._handle_done_future)

        @server_url_changed.connect
        def on_server_url_changed(sender, old_value, new_value):
            nonlocal instance
            if sender is not instance:
                return

            with background_ctx(LOGGER, type(self).__name__):
                if old_value == new_value:
                    return
                self.widget.api_url = new_value

        @default_instance_changed.connect
        def on_default_instance_changed(sender, old_instance, new_instance):
            nonlocal instance
            if instance is new_instance:
                return
            instance = new_instance
            with background_ctx(LOGGER, type(self).__name__):
                connect(instance)

        @channels_fetch_state_changed.connect
        def on_channels_state_changed(sender: Channels, old_value, new_value):
            LOGGER.debug(
                "Channels state update: %s %s %s", sender, old_value, new_value
            )
            nonlocal instance
            if sender.instance is not instance:
                return
            with background_ctx(LOGGER, type(self).__name__):
                self.widget.channels_state = new_value
                if new_value == ChannelsState.Loaded:
                    self.widget.channels = [channel.to_dict() for channel in sender]

        def unbind():
            nonlocal disconnected
            if disconnected:
                return
            channels_fetch_state_changed.disconnect(on_channels_state_changed)
            default_channel_changed.disconnect(on_channel_changed)
            access_token_changed.disconnect(on_access_token_changed)
            default_instance_changed.disconnect(on_default_instance_changed)
            server_url_changed.disconnect(on_server_url_changed)
            disconnected = True

        return unbind

    def _observe_notebook(self) -> Callable[[], None]:
        @notebook_name_changed.connect_via(self.info_watcher)
        def handle_name_change(sender, old_value, new_value):
            self.widget.notebook_name = new_value or "<unknown>"

        def unobserve():
            notebook_name_changed.disconnect(handle_name_change, self.info_watcher)

        return unobserve

    async def _observe_auth(self) -> None:
        while True:
            instance = get_instance()

            if not instance.access_token:
                await asyncio.wait(
                    [
                        wait_for_signal(access_token_changed, instance),
                        wait_for_signal(default_instance_changed),
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                continue

            try:
                access_token = await instance._get_access_token_async()
            except exc.NotLoggedIn:
                access_token = None

            if access_token:
                try:
                    await api.get_current_app(  # type: ignore
                        async_req=True,
                        _headers={"Authorization": f"Bearer {access_token}"},
                    )
                except ApiException as err:
                    if 400 <= err.status < 500:
                        LOGGER.warn(
                            "Logging out because of %d response to get_current_app",
                            err.status,
                        )
                        instance.logout()
            elif self.widget.auth_state == AuthState.Authenticated:
                LOGGER.warn(
                    "Logging out because of no access token",
                )
                instance.logout()

            await asyncio.sleep(900)

    @contextlib.contextmanager
    def _session_ctx(self, loop: asyncio.AbstractEventLoop):
        instance = get_instance()

        lock = asyncio_lock(loop=loop)

        session: Optional[SessionResponse] = None

        connect_task: Optional[asyncio.Task] = None

        queue = asyncio_queue(loop=loop)

        def date_or_null(value):
            if not value:
                return None
            return date_from_millis(value).isoformat()

        def end_session():
            if session is None:
                return
            instance.end_session(session.session_id)

        async def sender(ws: WebSocket):
            while True:
                await queue.get()
                message_count = 1

                while not queue.empty():
                    queue.get_nowait()
                    message_count += 1

                LOGGER.info(
                    "Started: %s (%s), finished: %s (%s)",
                    self.widget.jupyter_cell_started_at,
                    date_or_null(self.widget.jupyter_cell_started_at),
                    self.widget.jupyter_cell_finished_at,
                    date_or_null(self.widget.jupyter_cell_finished_at),
                )

                message = {
                    "url": self.widget.url,
                    "notebookName": self.widget.notebook_name,
                    "shellState": self.widget.jupyter_state.value,
                    "cellState": self.widget.jupyter_cell_state.value,
                    "cellText": self.widget.jupyter_cell_text,
                    "cellError": self.widget.jupyter_cell_error,
                    "executionNum": self.widget.jupyter_execution_num,
                    "cellStartedAt": date_or_null(self.widget.jupyter_cell_started_at),
                    "cellFinishedAt": date_or_null(
                        self.widget.jupyter_cell_finished_at
                    ),
                    "notifyOn": self.widget.monitoring_state,
                    "notifyChannel": self.widget.selected_channel,
                }
                LOGGER.debug("Sending ws message: %s", message)
                await ws.send(message)
                LOGGER.debug("Sent ws message")

                for _ in range(message_count):
                    queue.task_done()

        async def receiver(ws: WebSocket):
            async for item in ws:
                if not item["ok"]:
                    LOGGER.error("Error in websocket: %s", item)
                    continue

                msg_type = item["message"]["type"]
                if msg_type == "update":
                    LOGGER.debug("Update from session: %s", item)
                    state = item["message"]["session"]["state"]
                    if self.widget.monitoring_state != state["notifyOn"]:
                        self.widget.monitoring_state = state["notifyOn"]
                    if state.get(
                        "notifyChannel"
                    ) and self.widget.selected_channel != state.get("notifyChannel"):
                        self.widget.selected_channel = state.get("notifyChannel")
                elif msg_type == "action":
                    action = item["message"]["action"]
                    if action == "interrupt":
                        os.kill(os.getpid(), signal.SIGINT)
                    else:
                        LOGGER.error("Invalid action: %s", action)
                else:
                    LOGGER.warn("Unhandled ws message type: %s", msg_type)

        async def connect():
            with background_ctx(LOGGER, type(self).__name__):
                if session is None:
                    raise RuntimeError("No session created yet")

                try:
                    async with instance.session_connect(
                        session.session_id, False
                    ) as ws:
                        await asyncio.gather(sender(ws), receiver(ws))
                finally:
                    await instance.end_session(session.session_id, async_req=True)

        def handle_change(info):
            async def handle():
                nonlocal session, connect_task

                with background_ctx(LOGGER, type(self).__name__):
                    if self.widget.auth_state != AuthState.Authenticated:
                        return
                    if self.widget.notebook_name is None:
                        return

                    async with lock:
                        LOGGER.debug("Acquired lock")
                        if session is not None:
                            await queue.put(info)
                            return

                        session = await instance.create_session(
                            self.widget.notebook_name,
                            state=JupyterSessionState(
                                type="jupyter",
                                url=self.widget.url,
                                notebookName=self.widget.notebook_name,
                                shellState=self.widget.jupyter_state.value,
                                cellError=self.widget.jupyter_cell_error,
                                executionNum=self.widget.jupyter_execution_num,
                                cellStartedAt=date_or_null(
                                    self.widget.jupyter_cell_started_at
                                ),
                                cellFinishedAt=date_or_null(
                                    self.widget.jupyter_cell_finished_at
                                ),
                                notifyOn=self.widget.monitoring_state,
                                notifyChannel=self.widget.selected_channel,
                            ),
                            async_req=True,
                        )
                        LOGGER.info("Created session: %s", session.session_id)
                        self.widget.session = session.to_dict()

                        connect_task = asyncio.create_task(connect())
                        connect_task.add_done_callback(self._handle_done_future)

                        atexit.register(end_session)

            task = loop.create_task(handle())
            task.add_done_callback(self._handle_done_future)

        watch_properties = [
            "auth_state",
            "notebook_name",
            "url",
            "jupyter_state",
            "jupyter_cell_state",
            "jupyter_cell_error",
            "jupyter_cell_started_at",
            "jupyter_cell_finished_at",
            "monitoring_state",
            "selected_channel",
        ]

        self.widget.observe(handle_change, watch_properties)

        try:
            yield
        finally:
            if connect_task is not None:
                connect_task.cancel()

                with contextlib.suppress(asyncio.CancelledError):
                    loop.run_until_complete(connect_task)

            self.widget.unobserve(handle_change, watch_properties)

    async def main_loop(self) -> None:
        while not self.shutdown_event.is_set():
            await asyncio.sleep(0.1)

    def run(self) -> None:
        with contextlib.ExitStack() as stack:
            stack.enter_context(background_ctx(LOGGER, type(self).__name__))
            stack.enter_context(loop_ctx(self.loop))
            stack.enter_context(self._session_ctx(self.loop))

            loop = self.loop
            unobserve_widget = self._observe_widget(loop)
            unobserve_instance = self._observe_instance(loop)
            unobserve_jupyter = self._observe_jupyter(loop)
            unobserve_notebook = self._observe_notebook()

            tasks = []
            tasks.append(loop.create_task(self.history.main_loop()))
            tasks.append(loop.create_task(self.info_watcher.main_loop()))
            tasks.append(loop.create_task(self._observe_auth()))
            tasks.append(loop.create_task(observe_google_colab_url(self.widget)))

            try:
                LOGGER.info("Starting main loop")
                self.startup_event.set()
                loop.run_until_complete(self.main_loop())
            finally:
                LOGGER.debug("Exiting")
                for task in reversed(tasks):
                    with contextlib.suppress(asyncio.CancelledError):
                        task.cancel()
                unobserve_notebook()
                unobserve_jupyter()
                unobserve_instance()
                unobserve_widget()
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.close()


DEFAULT_WIDGET = None

DEFAULT_WIDGET_LOCK = threading.RLock()


def get_widget() -> LMKWidget:
    global DEFAULT_WIDGET
    with DEFAULT_WIDGET_LOCK:
        if DEFAULT_WIDGET is None:
            DEFAULT_WIDGET = LMKWidget()
        return DEFAULT_WIDGET


def set_widget(widget: LMKWidget) -> None:
    global DEFAULT_WIDGET
    with DEFAULT_WIDGET_LOCK:
        if widget is DEFAULT_WIDGET:
            return
        if DEFAULT_WIDGET is not None:
            DEFAULT_WIDGET.shutdown()
        DEFAULT_WIDGET = widget



import contextlib
import io
import logging
import os
import signal
import stat
from typing import Callable, Generator, List, Optional


LOGGER = logging.getLogger(__name__)


def socket_exists(path: str) -> bool:
    try:
        mode = os.stat(path).st_mode
    except FileNotFoundError:
        return False
    return stat.S_ISSOCK(mode)


@contextlib.contextmanager
def signal_handler_ctx(
    signals: List[int],
    handler: Callable[[int], None],
) -> Generator[None, None, None]:
    def handle_signal(signum, _):
        handler(signum)

    prev_handlers = {}
    for signal_num in signals:
        prev_handlers[signal_num] = signal.signal(signal_num, handle_signal)

    try:
        yield
    finally:
        for signal_num, old_handler in prev_handlers.items():
            signal.signal(signal_num, old_handler)


def read_last_lines(
    path: str, num_lines: int, max_size: Optional[int] = None
) -> List[str]:
    size = os.path.getsize(path)

    with open(path) as file:
        seek_start = max(size - max_size, 0) if max_size is not None else 0
        if seek_start > 0:
            file.seek(seek_start)
        lines = []
        for line in file:
            lines.append(line)
            if len(lines) > num_lines:
                lines.pop(0)

        return lines



from logging.config import dictConfig
from typing import Optional, IO, Dict, Any


LOG_FORMAT = "%(asctime)s [%(name)s - %(levelname)s] %(message)s"


def setup_logging(
    disable_existing: bool = True,
    level: str = "INFO",
    format: str = LOG_FORMAT,
    log_file: Optional[str] = None,
    log_stream: Optional[IO[str]] = None,
) -> None:
    handler_kwargs: Dict[str, Any] = {"class": "logging.StreamHandler"}
    if log_file is not None:
        handler_kwargs = {"class": "logging.FileHandler", "filename": log_file}
    elif log_stream is not None:
        handler_kwargs = {"class": "logging.StreamHandler", "stream": log_stream}

    config = {
        "version": 1,
        "disable_existing_loggers": disable_existing,
        "formatters": {
            "standard": {
                "format": format,
            }
        },
        "handlers": {
            "default": {"level": level, "formatter": "standard", **handler_kwargs}
        },
        "loggers": {
            "": {"handlers": ["default"], "level": "ERROR", "propagate": False},
            "lmk": {"handlers": ["default"], "level": level, "propagate": False},
        },
    }

    dictConfig(config)



from typing import List

from lmk.generated.models.notification_channel_response import (
    NotificationChannelResponse,
)


class LMKError(Exception):
    """
    Base class for errors raised by LMK
    """


class ConfigFileNotFound(LMKError):
    """
    Error indicating that a configuration file that was manually specified
    was not found.
    """

    def __init__(self, path: str) -> None:
        self.path = path
        super().__init__(f"Config file does not exist: {path}")


class NotLoggedIn(LMKError):
    """
    Error indicating that you are trying to do something that required authentication,
    but the current client is not logged in
    """

    def __init__(self) -> None:
        super().__init__("The current client is not logged in.")


class AuthSessionNotComplete(LMKError):
    """
    Error indicating that an authentication session
    has not been completed yet
    """

    def __init__(self, session_id: str) -> None:
        self.session_id = session_id
        super().__init__(f"Session not complete yet: {session_id}")


class AccessTokenAlreadyRetrieved(LMKError):
    """
    Error indicating that the access token for an auth
    session has already been retrieved
    """

    def __init__(self, session_id: str) -> None:
        self.session_id = session_id
        super().__init__(f"Access token already retrieved for session: {session_id}")


class AuthenticationTimeout(LMKError):
    """
    Error indicating that an interactive authentication session
    timed out
    """

    def __init__(self, session_id: str, timeout: float) -> None:
        self.session_id = session_id
        self.timeout = timeout
        super().__init__(
            f"Authentication timed out for session {session_id} after {timeout:.2f}s"
        )


class ChannelsNotFetched(LMKError):
    """
    Error indicating that you are trying to get channel information
    but it has not been fetched yet
    """

    def __init__(self, state: str) -> None:
        self.state = state
        super().__init__(f"Channels has not been fetched. State: {state}")


class MultipleChannelsMatched(LMKError):
    """
    Error indicating that get() was called on a Channels instance,
    but multiple channels matched query parameters
    """

    def __init__(self, channels: List[NotificationChannelResponse]) -> None:
        self.channels = channels
        channels_str = ", ".join([channel.name for channel in channels])
        super().__init__(f"Multiple channels matched parameters: {channels_str}")
